# -*- coding: utf-8 -*-
import numpy as np

from .configs import cutout_crop_rect_config, cutout_process_box_config
from .core import correct_background_boxes, generate_rect_coordinates
from .utils import generate_parameter, unpack_mm_params


class Cutout:
    """
    Description:
    CutOut class. This class allows crop part of image and fill this place some color.
    It also changes boxes and labels.

    Init Parameters:

    crop_rect_config (dict) - Config with parameters of cut and insert coordinates
    process_boxes_config (dict) - Config with parameters of maximum box area overlap and maximum box side overlap
    """

    def __init__(self, crop_rect_config: dict = None,
                 process_boxes_config: dict = None,):
        self.cr_conf = cutout_crop_rect_config if crop_rect_config is None else crop_rect_config
        self.pb_conf = cutout_process_box_config if process_boxes_config is None else process_boxes_config

        for def_key, def_val in cutout_crop_rect_config.items():
            if def_key not in self.cr_conf.keys() or self.cr_conf[def_key] is None:
                self.cr_conf[def_key] = def_val

        for def_key, def_val in cutout_process_box_config.items():
            if def_key not in self.pb_conf.keys() or self.pb_conf[def_key] is None:
                self.pb_conf[def_key] = def_val

    def apply(self,
              img: np.array,
              boxes: np.array,
              labels: np.array):
        """
        Description:
        CutOut function. It function crop part of image from second image and paste it into first image.
        Function also changes first and second boxes and labels.

        Parameters:

        img (np.array) - Background image. The image into which another image will be inserted.
        boxes (np.array) - Boxes for background image.
        labels (np.array) - Labels for background image.

        Return:
        (new_image, new_boxes, new_labels) - New image, boxes and labels
        """
        crop_min_x, crop_max_x = unpack_mm_params(self.cr_conf['crop_x'])
        crop_min_y, crop_max_y = unpack_mm_params(self.cr_conf['crop_y'])
        rect_min_h, rect_max_h = unpack_mm_params(self.cr_conf['rect_h'])
        rect_min_w, rect_max_w = unpack_mm_params(self.cr_conf['rect_w'])
        min_transp, max_transp = unpack_mm_params(self.cr_conf['transparency'])
        min_holes, max_holes = unpack_mm_params(self.cr_conf['hole_nums'])
        holes = generate_parameter(min_holes, max_holes)

        new_boxes = boxes.copy()
        new_labels = labels.copy()
        out_img = img.copy()

        for _ in range(holes):
            rect = generate_rect_coordinates(
                img_h=img.shape[0],
                img_w=img.shape[1],
                min_x=crop_min_x, min_y=crop_min_y,
                max_x=crop_max_x, max_y=crop_max_y,
                min_h=rect_min_h, min_w=rect_min_w,
                max_h=rect_max_h, max_w=rect_max_w,
            )

            x1, y1, x2, y2 = rect
            transp = generate_parameter(min_transp, max_transp)

            out_img[y1:y2, x1:x2:, :] = (out_img[y1:y2, x1:x2:, :] * transp).astype(np.uint8)

            if transp < self.pb_conf['transp_box_visibility']:
                new_boxes, new_labels = correct_background_boxes(
                    new_boxes, new_labels, rect,
                    self.pb_conf['max_overlap_area_ratio'],
                    self.pb_conf['min_height_result_ratio'],
                    self.pb_conf['min_width_result_ratio'],
                    self.pb_conf['max_height_intersection'],
                    self.pb_conf['max_width_intersection'],
                )

        return out_img, \
            np.array(new_boxes, dtype=np.float32), \
            np.array(new_labels)

    def __call__(self,
                 img: np.array,
                 boxes: np.array,
                 labels: np.array,):

        return self.apply(img, boxes, labels)



# -*- coding: utf-8 -*-
import numpy as np

from .configs import cutmix_crop_rect_config, cutmix_process_box_config
from .core import (correct_background_boxes, generate_rect_coordinates,
                   insert_image_in_background)
from .utils import unpack_mm_params


def shift_fg_rect_and_boxes(rect: np.array, boxes: np.array, shift_x: int, shift_y: int):
    """
    Description:
    The shift of the boxes is necessary after the cut image is inserted into
    a new random place on the background.
    Due to the fact that the starting and inserted places do not match,
    it is necessary to shift the boxes.

    Parameters:
    rect (np.array) - image rect coordinates
    boxes (np.array) - set of box coordinates
    shift_x (int) - X-axis shift
    shift_y (int) - Y-axis shift

    Returns:
    (new_rect, new_boxes) (np.array, np.array) - tuple of new image rectangle
        and box coordinates
    """
    x1, y1, x2, y2 = rect

    new_rect = np.array([
        x1 + shift_x,
        y1 + shift_y,
        x2 + shift_x,
        y2 + shift_y,
    ])

    new_boxes = boxes.copy()
    new_boxes[:, 0] += shift_x
    new_boxes[:, 1] += shift_y
    new_boxes[:, 2] += shift_x
    new_boxes[:, 3] += shift_y

    return new_rect, new_boxes


def correct_foreground_boxes(fg_boxes: np.array,
                             fg_labels: np.array,
                             image_rect: np.array,
                             max_overlap_area_ratio: np.array,
                             min_height_ratio: np.array,
                             min_width_ratio: np.array):
    """
    Description:
    Foreground boxes correction by custom parameters.

    Parameters:
    fg_boxes (np.array) - foreground boxes coordinates
    fg_labels (np.array) - foreground labels of boxes
    image_rect (np.array) - coordinates of rectangle that will be cutted from foreground
    max_overlap_area_ratio (float) - maximum box overlap threshold.
        If overlap is larger - box will be dropped
    min_height_ratio (float) - min part of height that should be exists or box will be dropped
    min_width_ratio (float) - min part of width that should be exists or box will be dropped

    Return:
    (new_box, critical_overlap_flag) - correct box coordinates and Boolean flag means
        that overlap exists and critical
    """
    x1, y1, x2, y2 = image_rect
    new_boxes = []
    new_labels = []

    # Поправляем боксы= вставленной картинки
    for i, (box, label) in enumerate(zip(fg_boxes, fg_labels)):
        f_x1, f_y1, f_x2, f_y2 = box
        start_box_height = f_y2 - f_y1
        start_box_width = f_x2 - f_x1

        l_side_in_rect = (f_x1 >= x1 and f_x1 <= x2)
        r_side_in_rect = (f_x2 >= x1 and f_x2 <= x2)
        t_side_in_rect = (f_y1 >= y1 and f_y1 <= y2)
        b_side_in_rect = (f_y2 >= y1 and f_y2 <= y2)

        # если какая-то часть бокса входит в новый прямоугольник
        if l_side_in_rect or r_side_in_rect or t_side_in_rect or b_side_in_rect:
            start_area = (f_x2 - f_x1) * (f_y2 - f_y1)
            # обрезаем части, которые вылезли за границу
            new_box = np.array([max(f_x1, x1), max(f_y1, y1), min(f_x2, x2), min(f_y2, y2)])

            end_box_height = new_box[3] - new_box[1]
            end_box_width = new_box[2] - new_box[0]
            end_box_area = end_box_height * end_box_width

            overlap_part = (start_area - end_box_area) / start_area

            if (end_box_height / start_box_height) < min_height_ratio:
                continue
            if (end_box_width / start_box_width) < min_width_ratio:
                continue
            if overlap_part > max_overlap_area_ratio:
                continue

            new_boxes.append(new_box)
            new_labels.append(label)
    return new_boxes, new_labels


class Cutmix:
    """
    Description:
    CutMix class. This class allows crop part of image from second image and paste it into first image.
    It also changes first and second boxes and labels.

    Init Parameters:

    crop_rect_config (dict) - Config with parameters of cut and insert coordinates
    process_boxes_config (dict) - Config with parameters of maximum box area overlap and maximum box side overlap
    """

    def __init__(self, crop_rect_config: dict = None,
                 process_boxes_config: dict = None,):
        self.cr_conf = cutmix_crop_rect_config if crop_rect_config is None else crop_rect_config
        self.pb_conf = cutmix_process_box_config if process_boxes_config is None else process_boxes_config

        for def_key, def_val in cutmix_crop_rect_config.items():
            if def_key not in self.cr_conf.keys() or self.cr_conf[def_key] is None:
                self.cr_conf[def_key] = def_val

        for def_key, def_val in cutmix_process_box_config.items():
            if def_key not in self.pb_conf.keys() or self.pb_conf[def_key] is None:
                self.pb_conf[def_key] = def_val

    def apply(self,
              bg_img: np.array,
              bg_boxes: np.array,
              bg_labels: np.array,

              fg_img: np.array,
              fg_boxes: np.array,
              fg_labels: np.array):
        """
        Description:
        Cutmix function. It function crop part of image from second image and paste it into first image.
        Function also changes first and second boxes and labels.

        Parameters:

        bg_img (np.array) - Background image. The image into which another image will be inserted.
        bg_boxes (np.array) - Boxes for background image.
        bg_labels (np.array) - Labels for background image.

        fg_img (np.array) - The image from which the rectangle will be cut
        fg_boxes (np.array) - Boxes for foreground image.
        fg_labels (np.array) - Labels for foreground image.

        Return:
        (new_image, new_boxes, new_labels) - New image, boxes and labels
        """
        crop_min_x, crop_max_x = unpack_mm_params(self.cr_conf['crop_x'])
        crop_min_y, crop_max_y = unpack_mm_params(self.cr_conf['crop_y'])
        rect_min_h, rect_max_h = unpack_mm_params(self.cr_conf['rect_h'])
        rect_min_w, rect_max_w = unpack_mm_params(self.cr_conf['rect_w'])

        insert_min_x, insert_max_x = unpack_mm_params(self.cr_conf['insert_x'])
        insert_min_y, insert_max_y = unpack_mm_params(self.cr_conf['insert_y'])

        img_h, img_w, _ = bg_img.shape

        fg_rect = generate_rect_coordinates(
            img_h=min(bg_img.shape[0], fg_img.shape[0]),
            img_w=min(bg_img.shape[1], fg_img.shape[1]),
            min_x=crop_min_x, min_y=crop_min_y,
            max_x=crop_max_x, max_y=crop_max_y,
            min_h=rect_min_h, min_w=rect_min_w,
            max_h=rect_max_h, max_w=rect_max_w,
        )

        x1, y1, x2, y2 = fg_rect
        cropped_img = fg_img[y1:y2, x1:x2]

        out_img, (shift_x, shift_y) = insert_image_in_background(
            bg_img, cropped_img, x1, y1,
            min_x=insert_min_x, min_y=insert_min_y,
            max_x=insert_max_x, max_y=insert_max_y,
        )

        fg_rect, fg_boxes = shift_fg_rect_and_boxes(fg_rect, fg_boxes, shift_x, shift_y)

        new_bg_boxes, new_bg_labels = correct_background_boxes(
            bg_boxes, bg_labels, fg_rect,
            self.pb_conf['max_overlap_area_ratio'],
            self.pb_conf['min_height_result_ratio'],
            self.pb_conf['min_width_result_ratio'],
            self.pb_conf['max_height_intersection'],
            self.pb_conf['max_width_intersection'],
        )

        new_fg_boxes, new_fg_labels = correct_foreground_boxes(
            fg_boxes, fg_labels, fg_rect,
            self.pb_conf['max_overlap_area_ratio'],
            self.pb_conf['min_height_result_ratio'],
            self.pb_conf['min_width_result_ratio'],
        )

        return out_img, \
            np.array(new_bg_boxes + new_fg_boxes, dtype=np.float32), \
            np.array(new_bg_labels + new_fg_labels)

    def __call__(self,
                 bg_img: np.array,
                 bg_boxes: np.array,
                 bg_labels: np.array,

                 fg_img: np.array,
                 fg_boxes: np.array,
                 fg_labels: np.array):

        return self.apply(bg_img, bg_boxes, bg_labels, fg_img, fg_boxes, fg_labels)



# -*- coding: utf-8 -*-
import numpy as np
import pytest


from augmixations.cutmix import (
    shift_fg_rect_and_boxes,
    correct_foreground_boxes,
    Cutmix,
)


@pytest.mark.parametrize('params',
                         [(np.array([0, 0, 200, 300]),
                           np.array([np.array([10, 10, 50, 50]),
                                     np.array([20, 10, 60, 40])]), 50, 100,
                             np.array([50, 100, 250, 400]),
                             np.array([np.array([60, 110, 100, 150]),
                                       np.array([70, 110, 110, 140])]))])
def test_shift_fg_rect_and_boxes(params):
    rect, boxes, shift_x, shift_y, result_rect, result_boxes = params
    new_rect, new_boxes = shift_fg_rect_and_boxes(rect, boxes, shift_x, shift_y)

    assert np.array_equal(result_rect, new_rect)
    assert np.array_equal(result_boxes, new_boxes)


@pytest.mark.parametrize('params', [
    # Кропаем боксы внутри прямоугольника
    (np.array([np.array([50, 50, 150, 150])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.9, 0.1, 0.1,
     np.array([np.array([100, 100, 150, 150])]), np.array(['1'], dtype=object)),

    (np.array([np.array([50, 150, 150, 250])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.9, 0.1, 0.1,
     np.array([np.array([100, 150, 150, 200])]), np.array(['1'], dtype=object)),

    (np.array([np.array([150, 150, 250, 250])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.9, 0.1, 0.1,
     np.array([np.array([150, 150, 200, 200])]), np.array(['1'], dtype=object)),

    (np.array([np.array([150, 50, 250, 150])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.9, 0.1, 0.1,
     np.array([np.array([150, 100, 200, 150])]), np.array(['1'], dtype=object)),

    # Удаляем боксы внутри прямоугольника по площади: max_overlap=0.5
    (np.array([np.array([50, 50, 150, 150])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.5, 0.1, 0.1,
     np.array([]), np.array([], dtype=object)),

    (np.array([np.array([50, 150, 150, 250])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.5, 0.1, 0.1,
     np.array([]), np.array([], dtype=object)),

    (np.array([np.array([150, 150, 250, 250])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.5, 0.1, 0.1,
     np.array([]), np.array([], dtype=object)),

    (np.array([np.array([150, 50, 250, 150])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.5, 0.1, 0.1,
     np.array([]), np.array([], dtype=object)),

    # Удаляем боксы вне области прямоугольника
    (np.array([np.array([50, 50, 75, 75])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.9, 0.1, 0.1,
     np.array([]), np.array([], dtype=object)),

    (np.array([np.array([50, 225, 75, 250])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.9, 0.1, 0.1,
     np.array([]), np.array([], dtype=object)),

    (np.array([np.array([225, 225, 250, 250])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.9, 0.1, 0.1,
     np.array([]), np.array([], dtype=object)),

    (np.array([np.array([225, 50, 250, 75])]), np.array(['1'], dtype=object),
     np.array([100, 100, 200, 200]), 0.9, 0.1, 0.1,
     np.array([]), np.array([], dtype=object)),

])
def test_correct_foreground_boxes(params):
    fg_boxes, fg_labels, rect, max_area, min_h, min_w, real_boxes, real_labels = params
    boxes, labels = correct_foreground_boxes(
        fg_boxes, fg_labels, rect, max_area, min_h, min_w)
    assert np.array_equal(real_boxes, boxes)
    assert np.array_equal(real_labels, labels)


# Чисто проверим, что метод работает)))
@pytest.mark.parametrize('params', [
    (np.ones((1000, 2000, 3), dtype=np.uint8)*255,
     np.array([np.array([50, 50, 150, 150])]),
     np.array(['1'], dtype=str),
     np.ones((600, 800, 3), dtype=np.uint8)*255,
     np.array([np.array([100, 100, 200, 200])]),
     np.array(['2'], dtype=str),
     {'crop_x': (75, 76), 'crop_y': (75, 76), 'rect_h': (150, 151), 'rect_w': (150, 151),
      'insert_x': (100, 101), 'insert_y': (100, 101), }, None,
     np.array([np.array([50, 50, 150, 150]), np.array([125, 125, 225, 225])]),
     np.array(['1', '2'], dtype=str)
     ),
    # Игнорируем боксы по минимальной высоте
    (np.ones((1000, 2000, 3), dtype=np.uint8)*255,
     np.array([np.array([50, 50, 150, 150])]),
     np.array(['1'], dtype=str),
     np.ones((600, 800, 3), dtype=np.uint8)*255,
     np.array([np.array([100, 100, 200, 200])]),
     np.array(['2'], dtype=str),
     {'crop_x': (75, 76), 'crop_y': (175, 176), 'rect_h': (150, 151),
      'rect_w': (150, 151), 'insert_x': (100, 101), 'insert_y': (100, 101), },
     {
        'max_overlap_area_ratio': 0.99,
         'min_height_result_ratio': 0.7,
        'min_width_result_ratio': 0.25,
        'max_height_intersection': 0.9,
        'max_width_intersection': 0.3,
    },
        np.array([]),
        np.array([]),
        # Игнорируем боксы по минимальной ширине
    ),
    (np.ones((1000, 2000, 3), dtype=np.uint8)*255,
     np.array([np.array([50, 50, 150, 150])]),
     np.array(['1'], dtype=str),
     np.ones((600, 800, 3), dtype=np.uint8)*255,
     np.array([np.array([100, 100, 200, 200])]),
     np.array(['2'], dtype=str),
     {'crop_x': (175, 176), 'crop_y': (75, 76),
      'rect_h': (150, 151), 'rect_w': (150, 151),
      'insert_x': (100, 101), 'insert_y': (100, 101), },
     {
        'max_overlap_area_ratio': 0.99,
        'min_height_result_ratio': 0.25,
        'min_width_result_ratio': 0.7,
        'max_height_intersection': 0.3,
        'max_width_intersection': 0.9,
    },
        np.array([]),
        np.array([]),
    ),
])
def test_cutmix(params):
    bg_img, bg_boxes, bg_labels, fg_img, fg_boxes, fg_labels, \
        crop_rect_config, process_boxes_config, real_boxes, real_labels = params

    cutmix = Cutmix(crop_rect_config, process_boxes_config)
    img, boxes, labels = cutmix(
        bg_img,
        bg_boxes,
        bg_labels,

        fg_img,
        fg_boxes,
        fg_labels,)

    assert np.array_equal(real_boxes, boxes)
    assert np.array_equal(real_labels, labels)


# Проверка, что метод не падает, когда некоторые параметры отсутствуют
@pytest.mark.parametrize('params', [
    (np.ones((1000, 2000, 3), dtype=np.uint8)*255,
     np.array([np.array([50, 50, 150, 150])]),
     np.array(['1'], dtype=str),
     np.ones((600, 800, 3), dtype=np.uint8)*255,
     np.array([np.array([100, 100, 200, 200])]),
     np.array(['2'], dtype=str),
     {'crop_y': (75, 76),
      'rect_h': (150, 151),
      'insert_x': (100, 101), },
     {
        'max_overlap_area_ratio': 0.99,
        'min_width_result_ratio': 0.7,
        'max_height_intersection': 0.3,
    },)
])
def test_cutmix_no_params(params):
    bg_img, bg_boxes, bg_labels, fg_img, fg_boxes, fg_labels, \
        crop_rect_config, process_boxes_config, = params

    cutmix = Cutmix(crop_rect_config, process_boxes_config)
    img, boxes, labels = cutmix(
        bg_img,
        bg_boxes,
        bg_labels,

        fg_img,
        fg_boxes,
        fg_labels,)

    assert img.shape == bg_img.shape



# -*- coding: utf-8 -*-
import pytest

from augmixations.utils import generate_parameter, unpack_mm_params


@pytest.mark.parametrize('params', [(1, 5), (0.1, 1.0), (5, 5)])
def test_generate_parameter(params):
    p1, p2 = params
    p = generate_parameter(p1, p2)

    assert p <= p2 and p >= p1


@pytest.mark.parametrize('params', [(1.0, 5), (1, 'Hello'), (None, 5)])
def test_generate_parameter_raise(params):
    p1, p2 = params
    with pytest.raises(Exception):
        _ = generate_parameter(p1, p2)


@pytest.mark.parametrize('params', [([1, 2], (1, 2)),
                                    ((2, 3), (2, 3)),
                                    (1, (1, 1)),
                                    (0.0, (0.0, 0.0))])
def test_unpack_mm_params(params):
    p, answer = params
    a = unpack_mm_params(p)

    assert answer == a


@pytest.mark.parametrize('params', [('Hello', None), (None, None)])
def test_unpack_mm_params_raise(params):
    p, answer = params
    with pytest.raises(Exception):
        _ = unpack_mm_params(p)



# -*- coding: utf-8 -*-
import numpy as np
import pytest

from augmixations.cutout import Cutout

# Проверка, что метод не падает


@pytest.mark.parametrize('params', [
    (np.ones((500, 500, 3), dtype=np.uint8)*255,
     np.array([np.array([50, 50, 150, 150])]),
     np.array(['1'], dtype=str),
     {
        'crop_x': (None, None),
        'crop_y': (None, None),
        'rect_h': (100, None),
        'rect_w': (100, None),
        'transparency': (0.0, 0.05),
        'hole_nums': (0, 3),
    },
        {
        'max_overlap_area_ratio': 0.75,
        'min_height_result_ratio': 0.25,
        'min_width_result_ratio': 0.25,
        'max_height_intersection': 0.9,
        'max_width_intersection': 0.9,
    },),
    (np.ones((500, 500, 3), dtype=np.uint8)*255,
     np.array([np.array([50, 50, 150, 150])]),
     np.array(['1'], dtype=str),
     {
        'crop_y': (None, None),
        'rect_h': (100, None),
        'rect_w': (100, None),
        'transparency': (0.0, 0.05),
        'hole_nums': (0, 3),
    },
        {
        'max_overlap_area_ratio': 0.75,
        'min_height_result_ratio': 0.25,
        'min_width_result_ratio': 0.25,
        'max_height_intersection': 0.9,
        'max_width_intersection': 0.9,
    },),
])
def test_cutout(params):
    img, boxes, labels, cr_config, pb_config = params

    cutout = Cutout(cr_config, pb_config)
    n_img, n_boxes, n_labels = cutout(img, boxes, labels)

    assert n_img.shape == img.shape



# -*- coding: utf-8 -*-
import numpy as np


def generate_parameter(min_p, max_p, name=None):
    """
    Description:
    Checking parameters type and return the same type random parameter if min_p and max_p are not equals.
    If min_p == max_p it returns min_p.

    Parameters:
    min_p (int or float) - min parameter value
    max_p (int or float) - max parameter value
    name (str) - parameter name, using for error messaage

    Returns:
    x (int or float) -  random value between min_p and max_p
    """

    if min_p == max_p:
        return min_p

    elif isinstance(min_p, float) and isinstance(max_p, float):
        return np.random.uniform(min_p, max_p)

    elif isinstance(min_p, int) and isinstance(max_p, int):
        return np.random.randint(min_p, max_p)

    else:
        raise Exception(f'Generate random parameter {name} error. Set both type of parameters int or float.')


def unpack_mm_params(p):
    """
    Description:
    Unpacking parameter that can contains min and max values or single value.
    Checking type of input parameter for correctly process.
    Parameters:
    p (int or float or list or tuple) - single value or min and max values for some range

    Returns:
    (p1, p2) (tuple of int or tuple of float) - tuple of min and max values for some range
    """
    if isinstance(p, (tuple, list)):
        return p[0], p[1]
    elif isinstance(p, (int, float)):
        return p, p
    else:
        raise Exception('Unknown input parameter type.')



# -*- coding: utf-8 -*-
import numpy as np
import pytest
from augmixations.mixup import Mixup


@pytest.mark.parametrize('params', [
    (np.ones((1000, 2000, 3), dtype=np.uint8)*255,
     np.array([np.array([50, 50, 150, 150])]),
     np.array(['1'], dtype=str),
     np.ones((1000, 2000, 3), dtype=np.uint8)*255,
     np.array([np.array([200, 200, 500, 500])]),
     np.array(['2'], dtype=str),
     np.array([np.array([50, 50, 150, 150]), np.array([200, 200, 500, 500])]),
     np.array(['1', '2'], dtype=str)
     ),
    (np.ones((1000, 2000, 3), dtype=np.uint8)*255,
     np.array([np.array([50, 50, 150, 150])]),
     np.array(['1'], dtype=str),
     np.ones((1000, 2000, 3), dtype=np.uint8)*255,
     np.array([np.array([200, 200, 500, 500])]),
     np.array(['2'], dtype=str),
     np.array([np.array([50, 50, 150, 150]), np.array([200, 200, 500, 500])]),
     np.array(['1', '2'], dtype=str),
     ),
    (np.ones((1000, 2000, 3), dtype=np.uint8)*255,
     np.array([np.array([50, 50, 150, 150])]),
     np.array(['2'], dtype=str),
     np.ones((1000, 2000, 3), dtype=np.uint8)*255,
     np.array([np.array([500, 500, 700, 700])]),
     np.array(['1'], dtype=str),
     np.array([np.array([50, 50, 150, 150]), np.array([500, 500, 700, 700])]),
     np.array(['2', '1'], dtype=str),
     ),
])
def test_mixup(params):
    bg_img, bg_boxes, bg_labels, fg_img, fg_boxes, fg_labels, \
        real_boxes, real_labels = params

    mixup = Mixup()
    img, boxes, labels = mixup(
        bg_img,
        bg_boxes,
        bg_labels,

        fg_img,
        fg_boxes,
        fg_labels,)

    assert np.array_equal(real_boxes, boxes)
    assert np.array_equal(real_labels, labels)
    assert bg_img.shape == img.shape
    assert fg_img.shape == img.shape


@pytest.mark.parametrize('params', [
    (np.ones((1000, 1000, 3), dtype=np.uint8)*255,
     np.array([np.array([50, 50, 150, 150])]),
     np.array(['1'], dtype=str),
     np.ones((1000, 2000, 3), dtype=np.uint8)*255,
     np.array([np.array([200, 200, 500, 500])]),
     np.array(['2'], dtype=str),
     ),
])
def test_mixup_diff_images(params):
    bg_img, bg_boxes, bg_labels, fg_img, fg_boxes, fg_labels = params

    mixup = Mixup()
    with pytest.raises(Exception):
        _, _, _ = mixup(
            bg_img,
            bg_boxes,
            bg_labels,

            fg_img,
            fg_boxes,
            fg_labels,)



# -*- coding: utf-8 -*-

cutout_crop_rect_config = {
    'crop_x': (None, None),  #  Min max
    'crop_y': (None, None),
    'rect_h': (None, None),
    'rect_w': (None, None),
    'transparency': (0.0, 0.05),
    'hole_nums': (1, 1),
}

cutout_process_box_config = {
    'transp_box_visibility': 0.2,

    'max_overlap_area_ratio': 0.75,

    'min_height_result_ratio': 0.25,
    'min_width_result_ratio': 0.25,

    'max_height_intersection': 0.9,
    'max_width_intersection': 0.9,
}

cutmix_crop_rect_config = {
    'crop_x': (None, None),  #  Min max
    'crop_y': (None, None),
    'rect_h': (None, None),
    'rect_w': (None, None),
    'insert_x': (None, None),
    'insert_y': (None, None),
}

cutmix_process_box_config = {
    'max_overlap_area_ratio': 0.75,

    'min_height_result_ratio': 0.25,
    'min_width_result_ratio': 0.25,

    'max_height_intersection': 0.9,
    'max_width_intersection': 0.9,
}



# -*- coding: utf-8 -*-
import numpy as np

from augmixations.utils import generate_parameter


def generate_rect_coordinates(img_h: int, img_w: int,
                              min_x: int = None, min_y: int = None,
                              max_x: int = None, max_y: int = None,
                              min_h: int = None, min_w: int = None,
                              max_h: int = None, max_w: int = None):
    """
    Description:
    Generation of coordinates by which the rectangle will be cut.

    Parameters:
    img_h (int), img_w (int) - Image parameters used to calculate the maximum value of a rectangle
    min_x (int), min_y (int) - Minimum allowed coordinates of the upper left point
    max_x (int), max_y (int) - Maximum allowed coordinates of the upper left point
    min_h (int), min_w (int) - Minimum allowed width and height of a rectangle
    max_h (int), max_w (int) - Maximum allowed width and height of a rectangle

    Returns:
    (x1, y1, x2, y2) (tuple of ints) -  Rectangle coordinates
    """

    min_h = 1 if min_h is None else min_h
    min_w = 1 if min_w is None else min_w

    max_h = img_h if max_h is None else max_h
    max_w = img_w if max_w is None else max_w

    rect_h = generate_parameter(min_h, max_h)
    rect_w = generate_parameter(min_w, max_w)

    min_x = 0 if min_x is None or min_x < 0 else min_x
    min_y = 0 if min_y is None or min_y < 0 else min_y

    max_x = img_w - rect_w if max_x is None else max_x
    max_y = img_h - rect_h if max_y is None else max_y

    x1 = generate_parameter(min_x, max_x)
    x2 = x1 + rect_w if x1 + rect_w <= img_w else img_w

    y1 = generate_parameter(min_y, max_y)
    y2 = y1 + rect_h if y1 + rect_h <= img_h else img_h

    return x1, y1, x2, y2


def insert_image_in_background(bg_img: np.array, rect_img: np.array,
                               start_x: int, start_y: int,
                               min_x: int = None, min_y: int = None,
                               max_x: int = None, max_y: int = None,):
    """
    Description:
    Paste the cropped image into a RANDOM place on the background.

    Parameters:
    bg_img (np.array) - Background image for paste another image into
    rect_img (np.array) - Image which will be insert into background image
    start_x (int), start_y (int) - Default left top coordinates of inserted image for
    calculate differece between crop and insert position
    min_x (int), min_y (int) - Minimum allowed coordinates of the upper left point
    max_x (int), max_y (int) - Maximum allowed width and height of a rectangle

    Returns:
    (image, (shift_x, shift_y)) (np.array, (int, int)) - tuple of image and
        coordinates of shift inserted image
    """

    bg_h, bg_w, _ = bg_img.shape
    r_h, r_w, _ = rect_img.shape

    min_x = 0 if min_x is None else min_x
    max_x = bg_w - r_w if max_x is None else max_x

    min_y = 0 if min_y is None else min_y
    max_y = bg_h - r_h if max_y is None else max_y

    x1 = np.random.randint(min_x, max_x)
    x1 = bg_w - r_w if x1 + r_w >= bg_w else x1
    x2 = x1 + r_w

    y1 = np.random.randint(min_y, max_y)
    y1 = bg_h - r_h if y1 + r_h >= bg_h else y1
    y2 = y1 + r_h

    shift_x = x1 - start_x
    shift_y = y1 - start_y

    image = bg_img.copy()
    image[y1:y2, x1:x2] = rect_img

    return image, (shift_x, shift_y)


def check_middle_part_overlap_critical(rect_info: dict,
                                       box_info: dict,
                                       max_overlap_area_ratio: float,
                                       max_height_intersection: float,
                                       max_width_intersection: float,
                                       debug: bool = False,
                                       label: object = None,):
    """
    Description:
    Check for overlap of some central part of one of the sides of the box.

    Parameters:
    rect_info (dict) - dictionary with parameters of a rectangle that is cut from another picture
    box_info (dict) - a dictionary with boxing parameters which is possibly overridden
    max_overlap_area_ratio (float) - maximum box overlap threshold
    debug (bool) - debug output flag
    label (object) - the label of the current box to display in debug information

    Return:
    Boolean flag means that overlap exists and critical
    """
    r_x1, r_y1 = rect_info['x1'], rect_info['y1']
    r_x2, r_y2 = rect_info['x2'], rect_info['y2']

    image_rect_height = rect_info['height']
    image_rect_width = rect_info['width']
    image_rect_area = rect_info['area']

    b_x1, b_y1 = box_info['x1'], box_info['y1']
    b_x2, b_y2 = box_info['x2'], box_info['y2']

    new_box = np.array([b_x1, b_y1, b_x2, b_y2])

    start_box_height = box_info['height']
    start_box_width = box_info['width']
    start_box_area = box_info['area']

    top_img_side_in_box = r_y1 >= b_y1 and r_y1 <= b_y2
    bot_img_side_in_box = r_y2 >= b_y1 and r_y2 <= b_y2
    right_img_side_in_box = r_x2 >= b_x1 and r_x2 <= b_x2
    left_img_side_in_box = r_x1 >= b_x1 and r_x1 <= b_x2

    top_img_side_higher_box = r_y1 <= b_y1
    bot_img_side_lower_box = r_y2 >= b_y2

    left_img_side_left_box = r_x1 <= b_x1
    right_img_side_right_box = r_x2 >= b_x2

    overlap_full_middle_vertical = right_img_side_in_box and left_img_side_in_box and \
        top_img_side_higher_box and bot_img_side_lower_box

    overlap_full_middle_horizontal = top_img_side_in_box and bot_img_side_in_box and \
        left_img_side_left_box and right_img_side_right_box

    overlap_part_middle_left = right_img_side_in_box and left_img_side_left_box and \
        top_img_side_in_box and bot_img_side_in_box

    overlap_part_middle_right = left_img_side_in_box and right_img_side_right_box and \
        top_img_side_in_box and bot_img_side_in_box

    overlap_part_middle_top = bot_img_side_in_box and top_img_side_higher_box and \
        left_img_side_in_box and right_img_side_in_box

    overlap_part_middle_bot = top_img_side_in_box and bot_img_side_lower_box and \
        left_img_side_in_box and right_img_side_in_box

    if top_img_side_in_box and bot_img_side_in_box and \
       right_img_side_in_box and left_img_side_in_box:
        overlap_part = image_rect_area / start_box_area
        if debug:
            print(f'{label} CRNTRAL MIDDLE OVERLAP')

    elif overlap_full_middle_vertical:
        overlap_area = start_box_height * image_rect_width
        overlap_part = overlap_area / start_box_area
        if debug:
            print(f'{label} FULL MIDDLE VERTICAL OVERLAP')

    elif overlap_full_middle_horizontal:
        overlap_area = start_box_width * image_rect_height
        overlap_part = overlap_area / start_box_area
        if debug:
            print(f'{label} FULL MIDDLE HORIZONTAL OVERLAP')

    elif overlap_part_middle_left:
        overlap_height = image_rect_height
        overlap_width = r_x2 - b_x1

        if overlap_height / start_box_height > max_height_intersection:
            new_box[0] = r_x2

        overlap_area = overlap_height * overlap_width
        overlap_part = overlap_area / start_box_area
        if debug:
            print(f'{label} PART MIDDLE LEFT HORIZONTAL OVERLAP')

    elif overlap_part_middle_right:
        overlap_height = image_rect_height
        overlap_width = b_x2 - r_x1

        if overlap_height / start_box_height > max_height_intersection:
            new_box[2] = r_x1

        overlap_area = overlap_height * overlap_width
        overlap_part = overlap_area / start_box_area
        if debug:
            print(f'{label} PART MIDDLE RIGHT HORIZONTAL OVERLAP')

    elif overlap_part_middle_top:
        overlap_height = r_y2 - b_y1
        overlap_width = image_rect_width

        if overlap_width / start_box_width > max_width_intersection:
            new_box[1] = r_y2

        overlap_area = overlap_height * overlap_width
        overlap_part = overlap_area / start_box_area
        if debug:
            print(f'{label} PART MIDDLE TOP VERTICAL OVERLAP')

    elif overlap_part_middle_bot:
        overlap_height = b_y2 - r_y1
        overlap_width = image_rect_width

        if overlap_width / start_box_width > max_width_intersection:
            new_box[3] = r_y1

        overlap_area = overlap_height * overlap_width
        overlap_part = overlap_area / start_box_area
        if debug:
            print(f'{label} PART MIDDLE BOT VERTICAL OVERLAP')
    else:
        return new_box, False

    return new_box, overlap_part > max_overlap_area_ratio


def correct_box_if_full_side_overlap(rect_info: dict, box_info: dict,
                                     max_overlap_area_ratio: float,
                                     debug: bool = False, label: object = None):
    """
    Description:
    Box verification for full side overlap case and correction it.
    Also checking overlap box area and return flag if overlap is critical.

    Parameters:
    rect_info (dict) - dictionary with parameters of a rectangle that is cut from another picture
    box_info (dict) - a dictionary with boxing parameters which is possibly overridden
    max_overlap_area_ratio (float) - maximum box overlap threshold
    debug (bool) - debug output flag
    label (object) - the label of the current box to display in debug information

    Return:
    (new_box, critical_overlap_flag) - correct box coordinates and Boolean flag means
        that overlap exists and critical
    """
    r_x1, r_y1 = rect_info['x1'], rect_info['y1']
    r_x2, r_y2 = rect_info['x2'], rect_info['y2']

    b_x1, b_y1 = box_info['x1'], box_info['y1']
    b_x2, b_y2 = box_info['x2'], box_info['y2']

    l_side_in_rect = (b_x1 >= r_x1 and b_x1 <= r_x2)
    r_side_in_rect = (b_x2 >= r_x1 and b_x2 <= r_x2)
    t_side_in_rect = (b_y1 >= r_y1 and b_y1 <= r_y2)
    b_side_in_rect = (b_y2 >= r_y1 and b_y2 <= r_y2)

    new_box = np.array([b_x1, b_y1, b_x2, b_y2])

    if l_side_in_rect and t_side_in_rect and b_side_in_rect:
        new_box[0] = r_x2
        if debug:
            print(f'{label} FULL LEFT SIDE OVERLAP')

    elif r_side_in_rect and t_side_in_rect and b_side_in_rect:
        new_box[2] = r_x1
        if debug:
            print(f'{label} FULL RIGHT SIDE OVERLAP')

    elif t_side_in_rect and r_side_in_rect and l_side_in_rect:
        new_box[1] = r_y2
        if debug:
            print(f'{label} FULL TOP SIDE OVERLAP')

    elif b_side_in_rect and r_side_in_rect and l_side_in_rect:
        new_box[3] = r_y1
        if debug:
            print(f'{label} FULL BOT SIDE OVERLAP')

    end_area = (new_box[3] - new_box[1]) * (new_box[2] - new_box[0])
    overlap_part = (box_info['area'] - end_area) / box_info['area']
    return new_box, overlap_part > max_overlap_area_ratio


def correct_box_if_some_alnge_overlap(rect_info: dict, box_info: dict,
                                      max_height_intersection: float,
                                      max_width_intersection: float,
                                      max_overlap_area_ratio: float,
                                      debug: bool = False, label: object = None):
    """
    Description:
    Box verification for some angle overlap case and correction it if some side overlap is critical.
    Also checking overlap box area and return flag if overlap is critical.

    Parameters:
    rect_info (dict) - dictionary with parameters of a rectangle that is cut from another picture
    box_info (dict) - a dictionary with boxing parameters which is possibly overridden
    max_height_intersection (float) - max part of height that should be exists or box will be changed
    max_width_intersection (float) - max part of width that should be exists or box will be changed
    max_overlap_area_ratio (float) - maximum box overlap threshold or box will be dropped
    debug (bool) - debug output flag
    label (object) - the label of the current box to display in debug information

    Return:
    (new_box, critical_overlap_flag) - correct box coordinates and Boolean flag means
        that overlap exists and critical
    """

    r_x1, r_y1 = rect_info['x1'], rect_info['y1']
    r_x2, r_y2 = rect_info['x2'], rect_info['y2']

    b_x1, b_y1 = box_info['x1'], box_info['y1']
    b_x2, b_y2 = box_info['x2'], box_info['y2']

    start_box_height = box_info['height']
    start_box_width = box_info['width']

    new_box = np.array([b_x1, b_y1, b_x2, b_y2])

    l_side_in_rect = (b_x1 >= r_x1 and b_x1 <= r_x2)
    r_side_in_rect = (b_x2 >= r_x1 and b_x2 <= r_x2)
    t_side_in_rect = (b_y1 >= r_y1 and b_y1 <= r_y2)
    b_side_in_rect = (b_y2 >= r_y1 and b_y2 <= r_y2)

    lt_in_rect = l_side_in_rect and t_side_in_rect
    lb_in_rect = l_side_in_rect and b_side_in_rect
    rt_in_rect = r_side_in_rect and t_side_in_rect
    rb_in_rect = r_side_in_rect and b_side_in_rect

    overlap_area = 0

    if lt_in_rect:
        if debug:
            print(f'{label} LEFT TOP OVERLAP')

        overlap_area = (r_x2 - b_x1) * (r_y2 - b_y1)

        tmp_height = new_box[3] - r_y2
        tmp_width = new_box[2] - r_x2

        bad_height = (1 - (tmp_height / start_box_height)) > max_height_intersection
        bad_width = (1 - (tmp_width / start_box_width)) > max_width_intersection

        if bad_height:
            new_box[0] = r_x2
        if bad_width:
            new_box[1] = r_y2

    elif rt_in_rect:
        if debug:
            print(f'{label} RIGHT TOP OVERLAP')

        overlap_area = (b_x2 - r_x1) * (r_y2 - b_y1)

        tmp_height = new_box[3] - r_y2
        tmp_width = r_x1 - new_box[0]

        bad_height = (1 - (tmp_height / start_box_height)) > max_height_intersection
        bad_width = (1 - (tmp_width / start_box_width)) > max_width_intersection

        if bad_height:
            new_box[2] = r_x1
        if bad_width:
            new_box[1] = r_y2

    elif rb_in_rect:
        if debug:
            print(f'{label} RIGHT BOT OVERLAP')

        overlap_area = (b_x2 - r_x1) * (b_y2 - r_y1)

        tmp_height = r_y1 - new_box[1]
        tmp_width = r_x1 - new_box[0]

        bad_height = (1 - (tmp_height / start_box_height)) > max_height_intersection
        bad_width = (1 - (tmp_width / start_box_width)) > max_width_intersection

        if bad_height:
            new_box[2] = r_x1
        if bad_width:
            new_box[3] = r_y1

    elif lb_in_rect:
        if debug:
            print(f'{label} LEFT BOT OVERLAP')

        overlap_area = (r_x2 - b_x1) * (b_y2 - r_y1)

        tmp_height = r_y1 - new_box[1]
        tmp_width = new_box[2] - r_x2

        bad_height = (1 - (tmp_height / start_box_height)) > max_height_intersection
        bad_width = (1 - (tmp_width / start_box_width)) > max_width_intersection

        if bad_height:
            new_box[1] = r_x2
        if bad_width:
            new_box[3] = r_y1

    end_box_area = box_info['area'] - overlap_area
    overlap_part = (box_info['area'] - end_box_area) / box_info['area']

    return new_box, overlap_part > max_overlap_area_ratio


def correct_background_boxes(bg_boxes: np.array,
                             bg_labels: np.array,
                             image_rect: np.array,
                             max_overlap_area_ratio: float,
                             min_height_ratio: float,
                             min_width_ratio: float,
                             max_height_intersection: float,
                             max_width_intersection: float,
                             debug: bool = False,
                             ):
    """
    Description:
    Background boxes correction by custom parameters.

    Parameters:
    bg_boxes (np.array) - background boxes coordinates
    bg_labels (np.array) - background labels of boxes
    image_rect (np.array) - coordinates of rectangle that will be inserted into background
    max_overlap_area_ratio (float) - maximum box overlap threshold. If overlap is larger - box will be dropped
    min_height_ratio (float) - min part of height that should be exists or box will be dropped
    min_width_ratio (float) - min part of width that should be exists or box will be dropped
    max_height_intersection (float) - max part of height that should be exists or box will be change
    max_width_intersection (float) - max part of width that should be exists or box will be change
    debug (bool) - debug output flag

    Return:
    (new_box, critical_overlap_flag) - correct box coordinates and Boolean flag means
        that overlap exists and critical
    """

    r_x1, r_y1, r_x2, r_y2 = image_rect

    rect_info = {
        'x1': r_x1,
        'y1': r_y1,
        'x2': r_x2,
        'y2': r_y2,
        'height': r_y2 - r_y1,
        'width': r_x2 - r_x1,
        'area': (r_y2 - r_y1) * (r_x2 - r_x1),
    }

    new_boxes = []
    new_labels = []

    for i, (box, label) in enumerate(zip(bg_boxes, bg_labels)):

        b_x1, b_y1, b_x2, b_y2 = box

        start_box_info = {
            'x1': b_x1,
            'y1': b_y1,
            'x2': b_x2,
            'y2': b_y2,
            'height': b_y2 - b_y1,
            'width': b_x2 - b_x1,
            'area': (b_y2 - b_y1) * (b_x2 - b_x1),
        }

        if b_x1 >= r_x1 and b_x2 <= r_x2 and b_y1 >= r_y1 and b_y2 <= r_y2:
            if debug:
                print(f'{label} TOTAL OVERLAP SKIP BY AREA')
            continue

        new_box, critical_overlap = check_middle_part_overlap_critical(
            rect_info, start_box_info,
            max_overlap_area_ratio,
            max_height_intersection,
            max_width_intersection,
            debug=debug,
            label=label)

        if critical_overlap:
            if debug:
                print('CRITICAL MIDLE PART OVERLAP')
            continue

        new_box, critical_overlap = correct_box_if_full_side_overlap(
            rect_info, start_box_info,
            max_overlap_area_ratio,
            debug=debug,
            label=label)

        if critical_overlap:
            if debug:
                print('CRITICAL FULL SIDE PART OVERLAP')
            continue

        new_box, critical_overlap = correct_box_if_some_alnge_overlap(
            rect_info, start_box_info,
            max_height_intersection,
            max_width_intersection,
            max_overlap_area_ratio,
            debug=debug,
            label=label)
        if critical_overlap:
            if debug:
                print('CRITICAL ANGLE PART OVERLAP')
            continue

        end_box_height = new_box[3] - new_box[1]
        end_box_width = new_box[2] - new_box[0]

        if (end_box_height / start_box_info['height']) < min_height_ratio:
            if debug:
                print('CRITICAL HEIGHT RATIO')
            continue
        if (end_box_width / start_box_info['width']) < min_width_ratio:
            if debug:
                print('CRITICAL WIDTH RATIO')
            continue

        new_boxes.append(new_box)
        new_labels.append(label)

    return new_boxes, new_labels



# -*- coding: utf-8 -*-
import os

from setuptools import setup, find_packages


def read(filename):
    with open(os.path.join(os.path.dirname(__file__), filename)) as f:
        file_content = f.read()
    return file_content


def get_version():
    """Get version from the package without actually importing it."""
    init = read('augmixations/__init__.py')
    for line in init.split('\n'):
        if line.startswith('__version__'):
            return eval(line.split('=')[1])


setup(
    name='augmixations',
    version=get_version(),
    description='Object detection augmentations.',
    long_description=read('README.txt'),
    long_description_content_type='text/markdown',
    url='https://github.com/TheDenk/augmixations',
    author='Karachev Denis',
    author_email='komedian@bk.ru',
    license='MIT',
    install_requires=['numpy>=1.11.1'],
    packages=find_packages(),
)



# -*- coding: utf-8 -*-
import numpy as np

from .utils import generate_parameter, unpack_mm_params


class Mixup:
    def __init__(self, transparency=(0.4, 0.6)):
        self.min_t, self.max_t = unpack_mm_params(transparency)

    def apply(self,
              bg_img: np.array,
              bg_boxes: np.array,
              bg_labels: np.array,

              fg_img: np.array,
              fg_boxes: np.array,
              fg_labels: np.array):
        if bg_img.shape != fg_img.shape:
            raise Exception('Both input images should be equal shape. \
            Current shapes: bg_img {bg_img.shape}, fg_img {fg_img.shape}')

        transp = generate_parameter(self.min_t, self.max_t, 'transparency')

        bg = bg_img.copy().astype(float)
        fg = fg_img.copy().astype(float)

        image = bg * transp + fg * (1 - transp)
        image = np.clip(image, 0, 255).astype(np.uint8)

        boxes = np.vstack([bg_boxes, fg_boxes])
        labels = np.hstack([bg_labels, fg_labels])

        return image, boxes, labels

    def __call__(self,
                 bg_img: np.array,
                 bg_boxes: np.array,
                 bg_labels: np.array,

                 fg_img: np.array,
                 fg_boxes: np.array,
                 fg_labels: np.array):

        return self.apply(bg_img, bg_boxes, bg_labels, fg_img, fg_boxes, fg_labels)
